-- keep wowace alive, donate!
local L = LibStub("AceLocale-3.0"):GetLocale("SpellReminder")

SpellReminder = LibStub('AceAddon-3.0'):NewAddon("SpellReminder", "AceConsole-3.0", "AceEvent-3.0", "AceTimer-3.0")
SpellReminder.sm = LibStub("LibSharedMedia-3.0")
function SpellReminder:FetchByIndex(smType,index)
	local list = self.sm:List(smType)
	return self.sm:Fetch(smType,list[index])
end
function SpellReminder:KeyByIndex(smType,index)
	local list = self.sm:List(smType)
	return list[index]
end

-- modules
local modulePrototype = {
	TrackSpell = function(self,eventName,spellName,targetName,targetGUID,timeLeft,filterType,icon,additional)
		local ci = SpellReminder:TrackSpell(eventName,spellName,targetName,targetGUID,timeLeft,filterType,icon,additional)
		if ci then ci.module = self:GetName() end
		return ci
	end,
--	OnEnable = function(self) print("OnEnable called!") end
}
SpellReminder:SetDefaultModulePrototype(modulePrototype)

local tooltip = CreateFrame("GameTooltip","SpellReminderTooltipScanning")
tooltip:SetOwner( WorldFrame, "ANCHOR_NONE" );
tooltip.LL, tooltip.LR = {}, {}
do
	local i
	for i=1,30 do
		tooltip.LL[i],tooltip.LR[i] = tooltip:CreateFontString(),tooltip:CreateFontString()
		tooltip:AddFontStrings(tooltip.LL[i],tooltip.LR[i]);
	end
	function tooltip:FindTipText(match,noleft,noright)
		for i=1,self:NumLines() do
			local l,r = self.LL[i]:GetText() or "" , self.LR[i]:GetText() or ""
			start,fin,txt = string.find(l,match)
			if start then return start,fin,txt end
		end
	end
	function tooltip:GetFullTipText(noleft,noright)
		local fulltext = ""
		for i=1,self:NumLines() do
			local l,r = self.LL[i]:GetText() or "" , self.LR[i]:GetText() or ""
			if not noleft then fulltext = fulltext.." "..l end
			if not noright then fulltext = fulltext.." "..r end
		end
		return fulltext
	end
end

local function sortFunc(a, b)
	local toReturn
	
	if a.isTimer ~= b.isTimer then
		toReturn = a.isTimer
	elseif a.value == b.value then
		toReturn = a.name > b.name
	else
		toReturn = a.value > b.value
	end
	
	local eithergrp = a:GetGroup() or b:GetGroup() -- both bars are not always set.. for some reason.
	if eithergrp then
		for i,anchor in pairs(SpellReminder.anchors) do
			if anchor.LBframe == eithergrp and not SpellReminder.db.profile.bargroups[i].reversesort then
				return not toReturn
			end
		end
	end
	return toReturn
end
local function ClipNumber(number, minimum, maximum, trueMax)
	return min(max(min(number,maximum),minimum),trueMax)
end
-- Thanks EBB
local roman_to_arabic = setmetatable({I = 1, V = 5, X = 10, L = 50, C = 100, D = 500, M = 1000}, {__index=function(self, roman)
	local arabic = 0
	local maxval = 0
	for i = roman:len(), 1, -1 do
		local digitval = self[roman:sub(i,i)]
		if digitval < maxval then
			arabic = arabic - digitval
		else
			arabic = arabic + digitval
			maxval = digitval
		end
	end
	self[roman] = arabic
	return arabic
end})

local _G = _G
local pairs = _G.pairs
local ipairs = _G.ipairs
local wipe = _G.wipe
local select = _G.select
local GetTime = _G.GetTime
local UnitName = _G.UnitName
local UnitGUID = _G.UnitGUID
local UnitBuff = _G.UnitBuff
local UnitDebuff = _G.UnitDebuff
local UnitIsDead = _G.UnitIsDead
local UnitExists = _G.UnitExists

local SPELLTYPE_NORMAL = ""
local SPELLTYPE_COOLDOWN = "CD"

local dbDefaultsProfile = {
	profile = {
		minimap = {
			hide = false,
		},
		ttg			= 0.3,
		disableEmphasis = false,
		lockbars	= false,
		debugmode	= false,
		-- output
		textWarnings	= 'None',
		audioWarningsEnabled = false,
		audioWarnings	= 'None',
		-- formatting
		textFormat	= "%t (%a) - %n",
		texture		= "Blizzard",
		baroverride	= false,
		barfont		= nil,
		barfontsize	= 11,
		barwidth	= 200,
		barheight	= 16,
		flashbars	= true,
		iconBorder	= true,
		-- spellsetup
		disableDetection	= false,
		defaultBarColor		= {["r"] = 1, ["g"] = 0, ["b"] = 0, ["a"] = 1},
		-- groups
		customGroups = {},
		bargroups = {
			['*'] = {
				position = {},
			},
		},
		--filters
		filterPrio = {'self','target','focus','buff','debuff','cooldown'},
		targetGroup_self	= 'normal',
		targetGroup_buff	= 'normal',
		targetGroup_debuff	= 'normal',
		targetGroup_cooldown= 'normal',
		targetGroup_focus	= 'normal',
		targetGroup_target	= 'normal',
		-- sink
		sinkOptions = {},
	},
}
local dbDefaultsChar = {
	char = {
		spellSetup	= {
			['**'] = {
				installed	= false,
				showSelf	= true,
				showOther	= true,
				singleTrack	= false,
				--fade		= 0,
			}
		},
	},
}

SpellReminder.hideWithoutStandby = true
SpellReminder.hasNoColor = true
SpellReminder.hasIcon = "Interface\\Icons\\Ability_Warrior_BattleShout"
SpellReminder.independentProfile = true

--SpellReminder.RegisteredBarGroups = {}

local spellSuffixes = {"",L[" Totem"],L[" Weapon"]}

SpellReminder.anchors = {}

local defaultBarGroup = { growup = false, emphasise=true, position = {}, scale = 1, }
--local castFailBuffer = 1
--local barRefreshBuffer = 0.5

-- stores information PER SPELL[Target]
SpellReminder.track = {
	casting = {},
	active = {},
}
local spellTracker = SpellReminder.track

function SpellReminder:OnInitialize()
	self.db = LibStub("AceDB-3.0"):New("SpellReminderDB",dbDefaultsProfile,true)
	self.chardb = LibStub("AceDB-3.0"):New("SpellReminderDBPerChar",dbDefaultsChar,true)
	if (self.chardb.global) then
		self:tmerge(self.chardb.global,self.chardb.char)
		wipe(self.chardb.global)
	end
	self.db.RegisterCallback(self, "OnProfileChanged", "ProfileChanged")
	self.db.RegisterCallback(self, "OnProfileCopied", "ProfileChanged")
	self.db.RegisterCallback(self, "OnProfileReset", "ProfileChanged")

	LibStub("LibBars-1.0"):Embed(self)
	LibStub("LibSink-2.0"):Embed(self)
	self:SetSinkStorage(self.db.profile.sinkOptions)
	self.options.args.general.args.output = self:GetSinkAce3OptionsDataTable()
	
	self.rootOptions = {
		type = "group",
		args = {
			general = self.options
		}
	}
	
	LibStub("AceConfig-3.0"):RegisterOptionsTable("SpellReminder", self.rootOptions, {'spellreminder','sr'})
	self.optFrames = {}
	self.optFrames.SpellReminder = LibStub("AceConfigDialog-3.0"):AddToBlizOptions("SpellReminder", "SpellReminder", nil,"general")
	self:RegisterModuleOptions("Profiles", LibStub("AceDBOptions-3.0"):GetOptionsTable(self.db), "Profiles")
	
	local ldb = LibStub:GetLibrary("LibDataBroker-1.1"):NewDataObject("SpellReminder", {
		type = "launcher",
		text = "SpellReminder",
		icon = "Interface\\Icons\\Ability_Warrior_BattleShout",
		OnClick = function(self, button)
			InterfaceOptionsFrame_OpenToCategory(SpellReminder.optFrames.Profiles) -- ensures that the submenu is opened
			InterfaceOptionsFrame_OpenToCategory(SpellReminder.optFrames.SpellReminder)
		end,
	})
	LibStub("LibDBIcon-1.0"):Register("SpellReminder", ldb, SpellReminder.db.profile.minimap)
end

function SpellReminder:RegisterModuleOptions(name, optTable, displayName)
	self.rootOptions.args[name] = optTable
	self.optFrames[name] = LibStub("AceConfigDialog-3.0"):AddToBlizOptions("SpellReminder", displayName or name, "SpellReminder", name)
end

function SpellReminder:COMBAT_LOG_EVENT(...)
	local eventName = select(3,...)
	if self.combatlog_events[eventName] ~= nil then
		self.combatlog_events[eventName](self,self:ProcessCombatLogArgs(...))
	end
end

function SpellReminder:AutoTanks(post)
	local groupHP, maxHP, minHP, groupCount, inRaid = 0,0,0,0,UnitInRaid("player")
	if inRaid then
		for i = 1, 40, 1 do
			if UnitExists("raid"..i) then
				local unitHP = UnitHealthMax("raid"..i)
				groupHP = groupHP + unitHP
				maxHP = max(maxHP,unitHP)
				minHP = min(minHP,unitHP)
				groupCount = groupCount + 1
			end
		end
	else
		local unitHP = UnitHealthMax("player")
		groupHP = groupHP + unitHP
		maxHP = max(maxHP,unitHP)
		minHP = min(minHP,unitHP)
		groupCount = groupCount + 1
		for i = 1, 4, 1 do
			if UnitExists("party"..i) then
				local unitHP = UnitHealthMax("party"..i)
				groupHP = groupHP + unitHP
				maxHP = max(maxHP,unitHP)
				minHP = min(minHP,unitHP)
				groupCount = groupCount + 1
			end
		end
	end
	local averageHP = groupHP / groupCount
	local threshold = averageHP + ((maxHP-averageHP) / 3)
	
	if inRaid then
		for i = 1, 40, 1 do
			if UnitExists("raid"..i) then
				local unitHP = UnitHealthMax("raid"..i)
				if unitHP >= threshold then
					self:Print("AutoTank: "..UnitName("raid"..i))
				end
			end
		end
	else
		local unitHP = UnitHealthMax("player")
		if unitHP >= threshold then
			if post then
						
			else
				self:Print("AutoTank: "..UnitName("player").." "..unitHP.." >= "..threshold)
			end
		end	
		for i = 1, 4, 1 do
			if UnitExists("party"..i) then
				local unitHP = UnitHealthMax("party"..i)
				if unitHP >= threshold then
					self:Print("AutoTank: "..UnitName("party"..i).." "..unitHP.." >= "..threshold)
				end
			end
		end
	end
end

function SpellReminder:OnEnable()
    -- Called when the addon is loaded
	self:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED","COMBAT_LOG_EVENT")
--	self:RegisterEvent("UNIT_SPELLCAST_SENT")
--	self:RegisterEvent("UNIT_SPELLCAST_SUCCEEDED")
--	self:RegisterEvent("UNIT_SPELLCAST_STOP")
--	self:RegisterEvent("UNIT_SPELLMISS")
--	self:RegisterEvent("UNIT_SPELLCAST_CHANNEL_STOP")
	self:RegisterEvent('UNIT_EXITED_VEHICLE',function()
		for e,castingInfo in pairs(spellTracker.active) do
			if castingInfo.tGUID == "VEHICLE" then
				self:CleanUp(eventName)--self:OnDurationEnded(eventName,true)
			end
		end
	end)
	self:RegisterEvent("PLAYER_DEAD")
	

	self:RegisterEvent("UNIT_SPELLCAST_CHANNEL_STOP",function(unit,spell,rank)
		local info = self.chardb.char.spellSetup[spell]
		if unit == "player" and info.installed then
			self:CleanUp(self:GetEventName(spell,UnitGUID("player")))
		end
	end)
	
	-- START
	self:RegisterCombatLogEvent('SPELL_CAST_SUCCESS')
	self:RegisterCombatLogEvent('UNIT_DIED')
	self:RegisterCombatLogEvent('PARTY_KILL','UNIT_DIED')
	self:RegisterCombatLogEvent('UNIT_DESTROYED','UNIT_DIED')
	
	self:ScheduleRepeatingTimer('RefreshAllBars',0.25)
	
	self:ProfileChanged()
end

function SpellReminder:ProfileChanged()
	-- cleanup obsolete info
	self.db.profile.barposition = nil
	self.db.profile.splitSelf = nil
	
	-- convert old info
		-- move bars from char to profile
		if self.chardb.char.bargroups ~= nil then
			for k,v in ipairs(self.chardb.char.bargroups) do
				if not self:AnchorExists(v.ident or v.name) then
					table.insert(self.db.profile.bargroups,v)
				end
			end
			self.chardb.char.bargroups = nil
		end
		-- update bar idents
		for k,v in ipairs(self.db.profile.bargroups) do
			self:RenameAnchor(k,v.ident or v.name)
		end
		for spellName, info in pairs(self.chardb.char.spellSetup) do
			-- update custom groups
			if info.customGroup then
				self.db.profile.customGroups[spellName] = info.customGroup
			end
			-- fix customSound
			if type(info.customSound) == 'number' then
				self.chardb.char.spellSetup[spellName].customSound = self:KeyByIndex('sound',info.customSound)
			end
		end
		-- fix audioWarnings
		if type(self.db.profile.audioWarnings) == 'number' then
			self.db.profile.audioWarnings = self:KeyByIndex('sound',self.db.profile.audioWarnings)
		end
		-- fix Bar Formatting
		if type(self.db.profile.texture) == 'number' then
			self.db.profile.texture = self:KeyByIndex('statusbar',self.db.profile.texture)
		end
		if type(self.db.profile.barfont) == 'number' then
			self.db.profile.barfont = self:KeyByIndex('font',self.db.profile.barfont)
		end
	-- end
	
	if not self:AnchorExists("normal") then
		self:AddAnchor('normal', { name=L["Normal"], position=(self.db.profile.barposition and self.db.profile.barposition["normal"]) or {}, })
	end
	if not self:AnchorExists("emphasis") then
		self:AddAnchor('emphasis', { name=L["Emphasis"], position=(self.db.profile.barposition and self.db.profile.barposition["emphasis"]) or {}, })
	end

	self:RefreshBarGroupList()
	self:RefreshGroupOptions()
	self:RefreshFilterPrioOptions()
	self:RefreshSpellOptions()
	self:CreateAnchors()
end

function SpellReminder:GetEventName(spellName,targetName)
	for k, spell in pairs(self.chardb.char.spellSetup) do
		if k == spellName then
			if spell.installed and spell.singleTrack == true then targetName = "" end
		end
	end
	return "SpellReminder_TRACKING_"..spellName..";"..targetName
end

function SpellReminder:GetSpellSetup(spellName,spellType)
	if spellType == SPELLTYPE_NORMAL then
		spellType = ''
	else
		spellType = spellType..';'
	end
	return self.chardb.char.spellSetup[spellType..spellName]
end

function SpellReminder:OnDisable()

end

function SpellReminder:AddAnchor(ident,overrideProperties)
	if self:AnchorExists(ident) then self:Print("Anchor with ID ("..ident..") already exists.") return end
	if not ident then self:Print('No anchor ID specified.') return end
	local newGroup = {}; self:tcopy(defaultBarGroup,newGroup)

	if overrideProperties then
		for k,v in pairs(overrideProperties) do newGroup[k] = v end
	end
	
	self.db.profile.bargroups[ident] = newGroup

	self:RefreshBarGroupList()
	self:RefreshGroupOptions()
	self:RefreshFilterPrioOptions()
	self:RefreshSpellOptions()
	self:CreateAnchors()
	
	return newGroup
end

function SpellReminder:RenameAnchor(oldident,newident)
	if not self:AnchorExists(oldident) then self:Print("Anchor with ID ("..oldident..") does not exist.") return end
	if self:AnchorExists(newident) then self:Print("Anchor with ID ("..newident..") already exists.") return end
	self:Print("Renaming "..oldident..' to '..newident)

	self.db.profile.bargroups[newident] = {}
	self:tcopy(self.db.profile.bargroups[oldident],self.db.profile.bargroups[newident])
	self.db.profile.bargroups[newident].name = newident

	for spell, group in pairs(self.db.profile.customGroups) do
		if group == oldident then
			self.db.profile.customGroups[spell] = newident
			self.chardb.char.spellSetup[spell].customGroup = nil
			self:Print(spell.." has had its custom group changed to "..newident..".")
		end
	end

	SpellReminder:DeleteAnchor(oldident)
	
	self:RefreshBarGroupList()
	self:RefreshGroupOptions()
	self:RefreshFilterPrioOptions()
	self:RefreshSpellOptions()
	self:CreateAnchors()
end

function SpellReminder:DeleteAnchor(id)
	-- reset spells
	for spell, group in pairs(self.db.profile.customGroups) do
		if group == id then
			self.chardb.char.spellSetup[spell].allowCustomGroup = false
			self.chardb.char.spellSetup[spell].customGroup = nil
			self.db.profile.customGroups[spell] = nil
			self:Print(L["%s has had its custom group reset."]:format(spell))
		end
	end
	
	-- reset target groups
	if self.db.profile.targetGroup_self		== id then self.db.profile.targetGroup_self = 'normal'; 	self:Print(L["Filter for '%s' has been reset."]:format("Self")) end
	if self.db.profile.targetGroup_buff		== id then self.db.profile.targetGroup_buff = 'normal'; 	self:Print(L["Filter for '%s' has been reset."]:format("Buffs")) end
	if self.db.profile.targetGroup_debuff	== id then self.db.profile.targetGroup_debuff = 'normal'; 	self:Print(L["Filter for '%s' has been reset."]:format("Debuffs")) end
	if self.db.profile.targetGroup_cooldown	== id then self.db.profile.targetGroup_cooldown = 'normal'; self:Print(L["Filter for '%s' has been reset."]:format("Cooldowns")) end
	if self.db.profile.targetGroup_focus	== id then self.db.profile.targetGroup_focus = 'normal'; 	self:Print(L["Filter for '%s' has been reset."]:format("Focus")) end
	if self.db.profile.targetGroup_target	== id then self.db.profile.targetGroup_target = 'normal';	self:Print(L["Filter for '%s' has been reset."]:format("Target")) end
	
	-- remove anchor
	if self.anchors[id] then
		self.anchors[id].LBframe:Hide()
		self.anchors[id] = nil
	end
	
	-- delete group settings
	self.db.profile.bargroups[id] = nil
	
	self:Print(L["You should reload your UI after deleting a bar group."])
	
	self:RefreshBarGroupList()
	self:RefreshGroupOptions()
	self:RefreshFilterPrioOptions()
	self:RefreshSpellOptions()
	self:CreateAnchors()
end

function SpellReminder:AnchorExists(ident)
	for i,a in pairs(SpellReminder.db.profile.bargroups) do
		if i == ident then return true end
	end
	return false
end

function SpellReminder:CreateAnchors()
	for i,v in pairs(self.anchors) do
		if not self.db.profile.bargroups[i] then
			self:DeleteAnchor(i)
		end
	end
	for ident,groupinfo in pairs(self.db.profile.bargroups) do
		if groupinfo == nil then
			self.db.profile.bargroups[ident] = nil
		else
			if self.anchors[ident] == nil then self.anchors[ident] = {} end
			self:tcopy(groupinfo,self.anchors[ident])
			-- new LibBars
			if self.anchors[ident].LBframe == nil then
				if self:GetBarGroup(groupinfo.name or ident) then
					self.anchors[ident].LBframe = self:GetBarGroup(groupinfo.name or ident)
				else
					self.anchors[ident].LBframe = self:NewBarGroup(groupinfo.name or ident, 1, self.db.profile.barwidth, self.db.profile.barheight, "ReminderFrame_"..ident)
					self.anchors[ident].LBframe.RegisterCallback(self,'AnchorMoved')
				end
			end
			local LBframe = self.anchors[ident].LBframe
			LBframe:Show()
			LBframe:SetWidth(self.db.profile.barwidth)
			LBframe.width = self.db.profile.barwidth
			LBframe:SetHeight(self.db.profile.barheight)
			LBframe.height = self.db.profile.barheight
			
			LBframe:SetTexture(self.sm:Fetch('statusbar',self.db.profile.texture))
			LBframe:SetFont(self.sm:Fetch('font',self.db.profile.barfont),self.db.profile.barfontsize)
			
			LBframe:SetSortFunction(sortFunc)
			LBframe:ClearAllPoints()
			if self.db.profile.bargroups[ident] and self.db.profile.bargroups[ident].position and self.db.profile.bargroups[ident].position.x and self.db.profile.bargroups[ident].position.y and self.db.profile.bargroups[ident].position.point and self.db.profile.bargroups[ident].position.anchor then
				LBframe:SetPoint(self.db.profile.bargroups[ident].position.point, UIParent, self.db.profile.bargroups[ident].position.anchor, self.db.profile.bargroups[ident].position.x, self.db.profile.bargroups[ident].position.y)
			else
				LBframe:SetPoint("CENTER", UIParent, "CENTER", 0, 0)
			end
			LBframe:SetScale(groupinfo.scale or 1)
			LBframe:ReverseGrowth((self.db.profile.bargroups[ident] and self.db.profile.bargroups[ident].growup) or false)
		end
	end
	self:ToggleAnchors()
end

function SpellReminder:ResetAnchors()
	if not self.db.profile.lockbars then self:ToggleAnchors() end
	for idx,anchor in pairs(self.anchors) do
		self:ResetAnchor(idx)
	end
	
	self:SaveAnchorPositions()
end

function SpellReminder:ResetAnchor(id)
	local anchor = self.anchors[id]
	anchor.LBframe:ClearAllPoints()
	anchor.LBframe:SetPoint("CENTER", 0, 50 / anchor.LBframe:GetScale())
end

function SpellReminder:ToggleAnchors()
	if self.db.profile.lockbars then
		for _,anchor in pairs(self.anchors) do
			anchor.LBframe:HideAnchor()
		end
	else
		for _,anchor in pairs(self.anchors) do
			anchor.LBframe:ShowAnchor()
		end
	end
end

function SpellReminder:AnchorMoved(cbk, group, x, y)
	self:SaveAnchorPositions()
end

function SpellReminder:SaveAnchorPositions()
	for i, anchorFrame in pairs(self.anchors) do
		local point, _, anchor, x, y = anchorFrame.LBframe:GetPoint()
		self.db.profile.bargroups[i].position.x = x --* s
		self.db.profile.bargroups[i].position.y = y --* s
		self.db.profile.bargroups[i].position.anchor = anchor
		self.db.profile.bargroups[i].position.point = point
	end
end

function SpellReminder:CreateAnchor(name, text, cRed, cGreen, cBlue)
	local f = CreateFrame("Button", "SpellReminder_"..name.."Anchor", UIParent)
	f:SetWidth(200)
	f:SetHeight(25)

	f.owner = self

	f:ClearAllPoints()
	
	if self.db.profile.bargroups[name] and self.db.profile.bargroups[name].position and self.db.profile.bargroups[name].position.x and self.db.profile.bargroups[name].position.y and self.db.profile.bargroups[name].position.point and self.db.profile.bargroups[name].position.anchor then
		f:SetPoint(self.db.profile.bargroups[name].position.point, UIParent, self.db.profile.bargroups[name].position.anchor, self.db.profile.bargroups[name].position.x, self.db.profile.bargroups[name].position.y)
	else
		f:SetPoint("CENTER", UIParent, "CENTER", 0, 50)--/s)
	end

	f:SetScript("OnDragStart", function() this:StartMoving() end )
	f:SetScript("OnDragStop",
		function()
			this:StopMovingOrSizing()
			self:SaveAnchorPositions()
        end)

	f:SetBackdrop({bgFile = "Interface/Tooltips/UI-Tooltip-Background",
                                            edgeFile = "Interface/Tooltips/UI-Tooltip-Border",
                                            tile = false, tileSize = 16, edgeSize = 16,
                                            insets = { left = 5, right =5, top = 5, bottom = 5 }})
	f:SetBackdropColor(cRed,cGreen,cBlue,.6)
	f:SetMovable(true)
	f:RegisterForDrag("LeftButton")

	f.Text = f:CreateFontString(nil, "OVERLAY")
	f.Text:SetFontObject(GameFontNormalSmall)
	f.Text:ClearAllPoints()
	f.Text:SetTextColor(1, 1, 1, 1)
	f.Text:SetWidth(200)
	f.Text:SetHeight(25)
	f.Text:SetPoint("TOPLEFT", f, "TOPLEFT")
	f.Text:SetJustifyH("CENTER")
	f.Text:SetJustifyV("MIDDLE")
	f.Text:SetText(text)

	f:Hide()

	return f
end

function SpellReminder:CreateTestBars()
	local i
	for i=1, 5 do
		if math.random()>0.5 then -- debuff
			self:TrackSpell("TEST_MANUAL_"..i,"Test "..i,"Enemy "..i,'NaN',i*5,"debuff",'Interface\\Icons\\INV_Misc_QuestionMark',{test=true,c={r=math.random(),g=math.random(),b=math.random()}})
		else
			local pseudoTarget = "Friendly "..i
			local pseudoTargetGUID = nil
			if math.random() > 0.5 then
				local pseudoTarget = "Player"
				local pseudoTargetGUID = UnitGUID('player')
			end
			self:TrackSpell("TEST_MANUAL_"..i,"Test "..i,pseudoTarget,pseudoTargetGUID,i*5,"buff",'Interface\\Icons\\INV_Misc_QuestionMark',{test=true,c={r=math.random(),g=math.random(),b=math.random()}})
		end
	end

	self:TrackSpell("TEST_FOCUS","Focus Test","Focus",'NaN',30,"focus",'Interface\\Icons\\INV_Misc_QuestionMark',{test=true,c={r=math.random(),g=math.random(),b=math.random()}})

	self:TrackSpell("TEST_TARGET","Target Test","Target",'NaN',30,"target",'Interface\\Icons\\INV_Misc_QuestionMark',{test=true,c={r=math.random(),g=math.random(),b=math.random()}})
	
	-- must reset options or AceConfig crashes on second execute
end

function SpellReminder:ClearSpells(clearAll)
	for spellName,info in pairs(self.chardb.char.spellSetup) do
		if (clearAll == true) or (clearAll == "inactive" and info.active == false) then
			self:ClearSpell(spellName,true)
		elseif (info.detected ~= true) then
			local found = false
			local i = 1
			while true do
				local spell, rank = GetSpellBookItemName(i, BOOKTYPE_SPELL)
				if (not spell) then	break end
				if spell == spellName then
					found = true
				end
				i = i + 1
			end
			
			if found == false then self:ClearSpell(spellName,true) end
		end
	end
	
	self:RefreshSpellOptions()
end

function SpellReminder:ClearSpell(spellName,dontRefresh)
	self.chardb.char.spellSetup[spellName] = nil
	self.spellOptions[spellName] = nil
	if not dontRefresh then self:RefreshSpellOptions() end
end

function SpellReminder:InstallWatch(spellName,rank,reminder,duration,icon)--,sharedName,reminder)
	if spellName == nil then return end
	if duration == nil then return end -- this can happen if the spell isnt available on this character
	duration = math.ceil(duration)
	
	local realSpellName = spellName
	if string.find(realSpellName,":") then
		_,_,_,realSpellName = string.find(realSpellName,"(.*):(.*)")
	end
		
	self:DebugPrint("Adding spell: "..realSpellName)
	self.chardb.char.spellSetup[spellName].installed = true
	if (self.chardb.char.spellSetup[spellName].active == nil) then
		self.chardb.char.spellSetup[spellName].active = true
	end
	self.chardb.char.spellSetup[spellName].displayName = self.chardb.char.spellSetup[spellName].displayName or realSpellName
	self.chardb.char.spellSetup[spellName].rank		= rank
	
	self.chardb.char.spellSetup[spellName].icon = icon or self:GetTexture(spellName)
	
	if self.chardb.char.spellSetup[spellName].d == nil or duration > self.chardb.char.spellSetup[spellName].d then
		self.chardb.char.spellSetup[spellName].d	= duration
		local defaultReminder = ClipNumber(math.ceil(duration*0.125),5,15,duration)--max(min(math.ceil(duration*0.125),15),5)
		self.chardb.char.spellSetup[spellName].r	= self.chardb.char.spellSetup[spellName].r or reminder or defaultReminder
	end
	
	self:RefreshSpellOptions()
	return self.chardb.char.spellSetup[spellName]
end

function SpellReminder:FlagIsSet(flags,flag)
	return bit.band(flags,flag) == flag
end

function SpellReminder:SPELL_CAST_SUCCESS(t)
	if not (self:FlagIsSet(t.sourceFlags,COMBATLOG_OBJECT_TYPE_PLAYER) and self:FlagIsSet(t.sourceFlags,COMBATLOG_OBJECT_AFFILIATION_MINE)) then return end
	if self:FlagIsSet(t.destFlags,COMBATLOG_OBJECT_NONE) then
		t.destGUID = t.sourceGUID
		t.destName = t.sourceName
		t.destFlags = t.sourceFlags
	end
	
	self:DebugPrint(string.format("SPELL_CAST_SUCCESS: %s -> %s : %s (%s~%s)",t.sourceName or '',t.destName or '',t.spellName or '',t.spellId or '',t.spellSchool or 0))
	local _,spellRank, icon = GetSpellInfo(t.spellId)
	if (not self.chardb.char.spellSetup[t.spellName] or not self.chardb.char.spellSetup[t.spellName].installed) then
		if self.db.profile.disableDetection == false then
			self:InstallWatch(t.spellName,"",nil,self:GetDuration(t.spellName),icon)
		else
			return
		end
	end
	local info = self.chardb.char.spellSetup[t.spellName]
	info.spellId = t.spellId
	if not info.installed or not info.active or not info.d or info.moduleManaged then return end
	
	if spellTracker.casting == nil then spellTracker.casting = {} end
	local bt = 'buff'
	if self:FlagIsSet(t.destFlags,COMBATLOG_OBJECT_REACTION_HOSTILE) then
		bt = 'debuff'
		self:SetSpellCategory(t.spellName,'Debuffs')
	else
		self:SetSpellCategory(t.spellName,'Buffs')
	end
	
	--local bt = ((destName == nil or self:FlagIsSet(destFlags,COMBATLOG_OBJECT_REACTION_HOSTILE)) and "buff") or "debuff" --bartype
	local eventName = self:GetEventName(t.spellName,t.destGUID)
	self:TrackSpell(eventName,t.spellName,t.destName,t.destGUID,info.d,bt,info.icon)
end
function SpellReminder:SPELL_MISSED(t)
	if not (self:FlagIsSet(t.sourceFlags,COMBATLOG_OBJECT_TYPE_PLAYER) and self:FlagIsSet(t.sourceFlags,COMBATLOG_OBJECT_AFFILIATION_MINE)) then return end
	local _,spellRank = GetSpellInfo(t.spellId)
	local eventName = self:GetEventName(t.spellName,t.destGUID)
--	for eventName,spell in pairs(spellTracker.casting) do
	--	if spell.n == t.spellName and spell.rank == spellRank then
			spellTracker.casting[eventName] = nil
--		end
--	end
--	for eventName,spell in pairs(spellTracker.active) do
--		if spell.n == t.spellName and spell.rank == spellRank then
--			self:OnDurationEnded(eventName,true)
--		end
--	end
end
function SpellReminder:UNIT_DIED(t)
	self:DebugPrint(t.event..": "..t.destName)
	for eventName,info in pairs(spellTracker.casting) do
		if info.tGUID == t.destGUID then
			self:CleanUp(eventName)
		end
	end
	for eventName,info in pairs(spellTracker.active) do
		if info.tGUID == t.destGUID then
			self:CleanUp(eventName)
		end
	end
end

function SpellReminder:IsTracking(eventName)
	return spellTracker.active[eventName] ~= nil
end

function SpellReminder:TrackSpell(eventName,spellName,targetName,targetGUID,timeLeft,filterType,icon,additional)
	if timeLeft < 0.5 then return end
	local setupInfo = self.chardb.char.spellSetup[spellName]
	if (spellTracker.active[eventName] == nil) then
		spellTracker.active[eventName] = {}
		self:DebugPrint("Tracking: "..spellName)
		-- Announce
		if setupInfo.installed and setupInfo.announceCast then
			local spellName, spellTarget = setupInfo.displayName or spellName, (targetName and targetName ~= UnitName("player") and " ("..targetName..")") or ""
			local message = L["%s Has been cast.  Will expire in %d seconds!"]:format(spellName..spellTarget,math.floor(timeLeft))
			if setupInfo.announceSay then SendChatMessage(message, "SAY") end
			if setupInfo.announceYell then SendChatMessage(message, "YELL") end
			if setupInfo.announceParty then SendChatMessage(message, "PARTY") end
			if setupInfo.announceRaid then SendChatMessage(message, "RAID") end
		end
	end
	local castingInfo = spellTracker.active[eventName]
	
	-- args has prio on these
	castingInfo.n 			= spellName
--	castingInfo.rank		= spellRank
	castingInfo.t			= targetName
	castingInfo.tGUID		= targetGUID
	castingInfo.filterType	= filterType -- buff/debuff/target/focus/self

	-- duration, always use the largest
	if (setupInfo.installed and (timeLeft > setupInfo.d)) then
		setupInfo.d = math.ceil(timeLeft)
		self:RefreshSpellOptions()
	end
	local duration = setupInfo.d or timeLeft
	castingInfo.s			= math.floor(GetTime() - (((setupInfo.installed and setupInfo.d) or duration) - timeLeft))
	
	-- setup has prio on these
	castingInfo.displayName	= setupInfo.displayName or spellName
	castingInfo.d			= setupInfo.d or timeLeft
	castingInfo.c			= setupInfo.c -- setup or default
	castingInfo.r			= setupInfo.r or ClipNumber(math.ceil(castingInfo.d*0.125),5,15,castingInfo.d) -- setup OR clip duration
	castingInfo.icon		= icon
	
	if additional then self:tcopy(additional,castingInfo) end

	if ((not setupInfo.installed) and castingInfo.d ~= nil and castingInfo.d > 0) and (self.db.profile.disableDetection == false) and (castingInfo.test ~= true) then
		self.chardb.char.spellSetup[castingInfo.n] = self:InstallWatch(castingInfo.n,castingInfo.rank,nil,castingInfo.d,castingInfo.icon)
		if not self:IsInSpellbook(castingInfo.n) then
			self:Print("Detected new spell! ("..self.chardb.char.spellSetup[castingInfo.n].displayName..")")
			self.chardb.char.spellSetup[castingInfo.n].detected = true
		end
		setupInfo = self.chardb.char.spellSetup[castingInfo.n]
	end
	if (not setupInfo.installed) and (castingInfo.test ~= true) then return end

	if setupInfo.installed then
		if castingInfo.icon ~= setupInfo.icon then setupInfo.icon = castingInfo.icon end
		if castingInfo.module then setupInfo.moduleManaged = true end

		if castingInfo.d > setupInfo.d then setupInfo.d = castingInfo.d	end
		castingInfo.r = castingInfo.r or setupInfo.r
	
		if castingInfo.tGUID ~= "COOLDOWN" and setupInfo.installed then
			if (not setupInfo.showSelf and castingInfo.tGUID == UnitGUID("player")) or
			   (not setupInfo.showOther and castingInfo.tGUID ~= UnitGUID("player")) then
				self:CleanUp(eventName)
				return
			end
		end -- this spell is flagged as Self Only, but its being cast on another target, ignore bar creation and free active and casting tables for this event.

		local ls = setupInfo.last_seen
		setupInfo.last_seen = time()
		if (ls == nil or ls < (time() - (86400*7))) then
			SpellReminder:RefreshSpellOptions()
		end
	end
	
	local timeLeft = castingInfo.d - (GetTime() - castingInfo.s)
	local reminder = castingInfo.r or SpellReminder.chardb.char.spellSetup[castingInfo.n]["r"]
	if math.floor(timeLeft) > math.ceil(reminder) then
		spellTracker.active[eventName].endingProcessed = nil
	end

	return spellTracker.active[eventName]
end

function SpellReminder:RefreshAllBars()
	--self:DebugPrint('RefreshAllBars')
	-- add new bars
	for e,castingInfo in pairs(spellTracker.active) do
		local setupInfo = self.chardb.char.spellSetup[castingInfo.n]
		local timeLeft = castingInfo.d - (GetTime() - castingInfo.s)
	--	local alreadyExists, needsRefresh = self.candybar:IsCandyBarRegistered(e), false
--		if alreadyExists then
--			local _,totTime,elapsed = self.candybar:CandyBarStatus(e)
--			local barTimeLeft = totTime - (elapsed or 9999)
--			needsRefresh = (barTimeLeft <= timeLeft-barRefreshBuffer) or (barTimeLeft >= timeLeft+barRefreshBuffer) or (castingInfo.applications ~= nil) -- time left is > 1sec out of sync (must be 1second, since time() does not return ms precision)
--		end
		-- persistant ??????
		if setupInfo.moduleManaged and not castingInfo.module then
			self:OnDurationEnded(e,true)
		elseif (timeLeft <= 0) or (setupInfo.active ~= true and castingInfo.test ~= true) then
			self:OnDurationEnded(e,setupInfo.active ~= true)
		else--if castingInfo.bar == nil or castingInfo.bar.SRstartTime ~= castingInfo.s then-- (castingInfo.bar and castingInfo.bar.startTime ~= castingInfo.s) or forceRefresh then--if --[[(not alreadyExists) or ]]needsRefresh or (castingInfo.targetGroupId ~= targetGroup) or forceRefresh then
--			local resort = false
	--		if castingInfo.bar == nil or castingInfo.bar.SRstartTime ~= castingInfo.s then
	--			resort = true
	--		end
			if setupInfo.installed and setupInfo.c and setupInfo.c ~= castingInfo.c then castingInfo.c = {}; self:tcopy(setupInfo.c,castingInfo.c) end-- copy to prevent setup color from being recycled by reference
			self:CreateReminderBar(e,castingInfo,timeLeft)
			local reminder = castingInfo.r or SpellReminder.chardb.char.spellSetup[castingInfo.n]["r"] or ClipNumber(math.ceil(castingInfo.d*0.125),5,15,castingInfo.d)
			if timeLeft - reminder <= 0 then
				self:OnDurationEnding(e)
			end
		end
	end
	
	-- any which are still existing as bars, but are no longer active?
	for i, anchor in pairs(self.anchors) do
		local bars = anchor.LBframe:GetBars()
		if bars then for e,bar in pairs(bars) do
			if spellTracker.active[e] == nil then --and not bar.fading then
				bar:Fade()
			end
		end end
	end
end

function SpellReminder:FindTargetGroup(castingInfo)
	local targetGroupId = nil
	local setupInfo = self.chardb.char.spellSetup[castingInfo.n]
	
	if setupInfo.installed and setupInfo.allowCustomGroup and self.db.profile.customGroups[castingInfo.n] and self.anchors[self.db.profile.customGroups[castingInfo.n]] then
		return self.db.profile.customGroups[castingInfo.n]
	end
	
	for _,fltrName in ipairs(self.db.profile.filterPrio) do
		if (castingInfo.filterType == fltrName) or castingInfo.tGUID == UnitGUID((fltrName == "self" and "player") or fltrName) then
			return self.db.profile['targetGroup_'..fltrName]
		end
	end
	return 'normal'
end

function SpellReminder:CreateReminderBar(e, castingInfo, timeLeft, targetGroupId)
	local color = castingInfo.c or self.db.profile.defaultBarColor
	local duration = castingInfo.d
	local reminder = castingInfo.r or SpellReminder.chardb.char.spellSetup[castingInfo.n]["r"] or ClipNumber(math.ceil(castingInfo.d*0.125),5,15,castingInfo.d)
	local target = castingInfo.t or UnitName("player")

	if targetGroupId == nil then
		targetGroupId = self:FindTargetGroup(castingInfo)
		if timeLeft <= reminder and (self.db.profile.bargroups[targetGroupId] and self.db.profile.bargroups[targetGroupId].emphasise) then
            targetGroupId = 'emphasis'
		end
	end
	if (not targetGroupId) or (not self.anchors[targetGroupId]) then
		if castingInfo.bar then
			castingInfo.bar:Fade()
			castingInfo.bar = nil
		end
		return
	end
	
	castingInfo.targetGroupId = targetGroupId
	local targetLibBarGroup = self.anchors[targetGroupId].LBframe
	
	-- build name
	local text = castingInfo.customFormat or self.db.profile.textFormat
	text = string.gsub(text,"%%n",castingInfo.displayName or castingInfo.n)
	text = string.gsub(text,"%%r",castingInfo.rank or "")
	text = string.gsub(text,"%%t",target)
	if castingInfo.applications == 0 then castingInfo.applications = nil end
	text = string.gsub(text,"%%a",castingInfo.applications or "")
	text = string.gsub(text, "[(][)]", "") -- trim empty parenthesis
	text = string.gsub(text, "^%s*(.-)%s*$", "%1") -- trim whitespaces
	
	if castingInfo.bar and castingInfo.bar:GetGroup() and not targetLibBarGroup:GetBar(e) then
		local oldGroup = castingInfo.bar:GetGroup()
		oldGroup:MoveBarToGroup(castingInfo.bar,targetLibBarGroup)--RemoveBar(castingInfo.bar)
		oldGroup:SortBars()
		targetLibBarGroup:SortBars()
	end
	if (not castingInfo.bar) or (castingInfo.s ~= castingInfo.bar.SRstartTime) or (castingInfo.bar:GetLength() ~= self.db.profile.barwidth) or (castingInfo.bar:GetThickness() ~= self.db.profile.barheight) then
		local bar, isNew = targetLibBarGroup:NewTimerBar(e, text, timeLeft, duration, castingInfo.icon, (self.db.profile.flashbars and reminder) or 0)
		if isNew then
			self:DebugPrint("Created Bar: "..castingInfo.n..": duration="..(duration or "no duration")..": reminder="..(reminder or "no reminder")..": timeleft="..(timeLeft or "none"))
		end
		bar.SRstartTime = castingInfo.s
		bar.updateDelay = 0 -- update on every frame .  Jerky bars = bad
		castingInfo.bar = bar
		targetLibBarGroup:SortBars()
	end
	castingInfo.bar:SetText(text)
	castingInfo.bar:SetIcon(castingInfo.icon)
	castingInfo.bar:UnsetAllColors()
	castingInfo.bar:SetColorAt(0,color.r,color.g,color.b,1)
	if self.db.profile.iconBorder then
		castingInfo.bar.icon:SetTexCoord(0, 1, 0, 1)
	else -- crop off the border
		castingInfo.bar.icon:SetTexCoord(0.07, 0.93, 0.07, 0.93)
	end
end

function SpellReminder:PLAYER_DEAD()
	for eventName,info in pairs(spellTracker.casting) do
		if info ~= nil then
			self:CleanUp(eventName)
		end
	end
	for eventName,info in pairs(spellTracker.active) do
		if info ~= nil then
			self:CleanUp(eventName)
		end
	end
end

function SpellReminder:OnDurationEnding(eventName,quiet)
	local castingInfo = spellTracker.active[eventName]
	if not castingInfo then self:CleanUp(eventName) return end
	if castingInfo.endingProcessed == true then return end
	castingInfo.endingProcessed = true
	self:DebugPrint("OnDurationEnding: "..eventName)
	local spellID, spellName, spellTarget = castingInfo.n, castingInfo.displayName, (castingInfo.t and castingInfo.t ~= UnitName("player") and " ("..castingInfo.t..")") or ""
	local setupInfo = self.chardb.char.spellSetup[spellID]

	if (not quiet) and setupInfo and setupInfo.installed then
		if setupInfo.allowCustomSound and setupInfo.customSound then
			PlaySoundFile(SpellReminder.sm:Fetch('sound', setupInfo.customSound))
		elseif self.db.profile.audioWarningsEnabled then
			PlaySoundFile(SpellReminder.sm:Fetch('sound', self.db.profile.audioWarnings))
		end
		
		local timeLeft = castingInfo.d - (GetTime() - castingInfo.s)
		local message = L["%s over in %d seconds!"]:format(spellName..spellTarget,math.floor(timeLeft))
	
		-- UI Text Warning
		if (self.db.profile.textWarnings == 4 or self.db.profile.textWarnings == 2) then
			self:Pour(message,1,1,1,nil,nil,nil,true)
		end

		-- Announce
		if setupInfo.announceEnding then
			if setupInfo.announceSay then SendChatMessage(message, "SAY") end
			if setupInfo.announceYell then SendChatMessage(message, "YELL") end
			if setupInfo.announceParty then SendChatMessage(message, "PARTY") end
			if setupInfo.announceRaid then SendChatMessage(message, "RAID") end
		end
	end
end

function SpellReminder:OnDurationEnded(eventName,quiet)
	self:DebugPrint("OnDurationEnded: "..eventName)
	
	local castingInfo = spellTracker.active[eventName]
	local spellID, spellName, spellTarget = castingInfo.n, castingInfo.displayName, (castingInfo.t and castingInfo.t ~= UnitName("player") and " ("..castingInfo.t..")") or ""
	local setupInfo = self.chardb.char.spellSetup[spellID]

	if (not quiet) and setupInfo and setupInfo.installed then
		if setupInfo.allowCustomSound and setupInfo.customSound then
			PlaySoundFile(SpellReminder.sm:Fetch('sound', setupInfo.customSound))
		elseif self.db.profile.audioWarningsEnabled then
			PlaySoundFile(SpellReminder.sm:Fetch('sound', self.db.profile.audioWarnings))
		end
		
		local message = L["%s has EXPIRED!"]:format(spellName..spellTarget)
		
		-- UI Text Warning
		if (self.db.profile.textWarnings == 4 or self.db.profile.textWarnings == 3) then
			self:Pour(message,1,1,1,nil,nil,nil,true)
		end
		
		-- Announce
		if setupInfo.announceEnded then
			if setupInfo.announceSay then SendChatMessage(message, "SAY") end
			if setupInfo.announceYell then SendChatMessage(message, "YELL") end
			if setupInfo.announceParty then SendChatMessage(message, "PARTY") end
			if setupInfo.announceRaid then SendChatMessage(message, "RAID") end
		end
	end
	
	self:CleanUp(eventName)
end

function SpellReminder:CleanUp(eventName)
	if eventName == nil then return end

	spellTracker.casting[eventName] = nil
	spellTracker.active[eventName] = nil
end

--[[function SpellReminder:GetRegisteredGroup(eventName)
	local regdGroup = nil
	for k,v in pairs(self.anchors) do
		if self.candybar:IsCandyBarRegisteredWithGroup(eventName,"ReminderCandy_"..k) then
			regdGroup = k
			break
		end
	end
	return regdGroup
end]]

function SpellReminder:GetTexture(name)
	local i = 1
	while true do
		local nm = GetSpellBookItemName(i, BOOKTYPE_SPELL)
		if not nm then break
		elseif nm == name then
			return GetSpellTexture(i, BOOKTYPE_SPELL)
		end
		i = i + 1
	end
	-- still not found? try pets spellbook
	local i = 1
	while true do
		local nm = GetSpellBookItemName(i, BOOKTYPE_PET)
		if not nm then break
		elseif nm == name then
			return GetSpellTexture(i, BOOKTYPE_PET)
		end
		i = i + 1
	end
	
	if true then return "Interface\\Icons\\INV_Misc_QuestionMark" end -- we dont know the texture, use questionmark
	
	-- finally, do the talents
	local numTabs = GetNumTalentTabs();
	local t
	for t=1, numTabs do
	    local numTalents = GetNumTalents(t);
	    for i=1, numTalents do
	        local nameTalent, icon, tier, column, currRank, maxRank= GetTalentInfo(t,i);
			if nameTalent == spellName and currRank > 0 then
				local buff, debuff = self:FindDurationInText(tiptext)
				return buff, debuff, currRank
			end
	    end
	end
end

function SpellReminder:BarExists(spellName)
	for e, info in pairs(spellTracker.active) do
		if info.n == spellName then
			return true
		end
	end
	return false
end

function SpellReminder:GetPlayerBuffDuration(spellName)
	local i = 1
	while (UnitBuff("player",i) ~= nil) do
		local name, rank, iconTexture, count, duration, timeLeft =  UnitBuff("player", i)
		if name == spellName and duration ~= nil then return duration end
		i = i + 1
	end
	local i = 1
	while (UnitBuff("pet",i) ~= nil) do
		local name, rank, iconTexture, count, duration, timeLeft =  UnitBuff("pet", i)
		if name == spellName and duration ~= nil then return duration end
		i = i + 1
	end
end

function SpellReminder:IsInSpellbook(spellName)
	for _,suffix in ipairs(spellSuffixes) do
		local i = 1
		while true do
			local success, spell, rank = pcall(GetSpellBookItemName, i, BOOKTYPE_SPELL)
			if (not success) then break end
			if (not spell) then	break end
			if spell == spellName..suffix then
				return true
			end
			i = i + 1
		end
	end
	
	return false
end

function SpellReminder:GetTempBuffName(id)
	local rank
	tooltip:SetInventoryItem("player", id)
	local _, _, buffname = tooltip:FindTipText("^(.+) %(%d+ [^%)]+%)$")
	if buffname then
		buffname = string.gsub(buffname, " %(%d+ [^%)]+%)", "") -- remove 2nd brackets for buffs with charges
		local tname, trank = string.match(buffname, "(.*) (%d*)$")
		if tname then
			buffname = tname
			rank = trank
		else
			local tname, trank = string.match(buffname, "(.*) ([CDILMVX]*)$")
			if tname then
				buffname = tname
				rank = roman_to_arabic[trank]
			end
		end
		return buffname, rank
	end
	local itemlink = GetInventoryItemLink("player", id)
	if itemlink then
		local name = GetItemInfo(itemlink)
		return name or "Weapon "..id
	end
	return "Weapon "..id
end

function SpellReminder:GetDuration(spellName, chkRank)
	for _,suffix in pairs(spellSuffixes) do
		local toreturn = nil
		local lastRank = nil
		local i = 1
		while true do
			local success, spell, rank = pcall(GetSpellBookItemName, i, BOOKTYPE_SPELL)
			if (not success) then break end
			if (not spell) then	break end
			if spell == spellName..suffix then
				if (chkRank == nil and (lastRank == nil or rank > lastRank)) or (rank == chkRank) then
					lastRank = rank
					tooltip:SetSpellBookItem(i,BOOKTYPE_SPELL)
					local tiptext = tooltip:GetFullTipText()
					local buff = self:FindDurationInText(tiptext)
					toreturn = { buff, rank }
				end
			end
			i = i + 1
		end
		if toreturn ~= nil then return toreturn[1], toreturn[2] end
	end

	return nil
end

function SpellReminder:FindDurationInText(tiptext)
	local pre = {"for","over","lasts","lasting","up to"}
	local post = {"sec","min","hour"}
	local durationBuff = nil
	local durationDebuff = nil
	for _,preVal in ipairs(pre) do -- loop through prefixes
		preVal = string.lower(preVal)
		for _,postVal in ipairs(post) do  -- loop through suffixes
			postVal = string.lower(postVal)
			for newDur in string.gmatch(string.lower(tiptext),preVal.." (%d+) "..postVal) do
				newDur = tonumber(newDur)
				if newDur ~= nil then
					if postVal == "min" then
						newDur = newDur * 60
					elseif postVal == "hour" then
						newDur = newDur * 60 * 60
					end
					if (durationBuff == nil) then
						durationBuff = newDur
					elseif (newDur > durationBuff) then
						durationDebuff = durationBuff
						durationBuff = newDur
					end
				end
			end
		end
	end
	
	return tonumber(durationBuff), tonumber(durationDebuff)
end

function SpellReminder:DebugPrint(text)
	if not self.db.profile.debugmode then return end
	self:Print(text)
end

function SpellReminder:tcopy(from,to)   -- "to" must be a table (possibly empty)
	assert(type(from)=="table","Error copying table contents, source is not a table.")
	assert(type(to)  =="table","Error copying table contents, destination is not a table.")
	for k,v in pairs(from) do
		if(type(v)=="table") then
			if (type(to[k]~='table')) then to[k] = {} end
			self:tcopy(v,to[k])
		else
			to[k] = v
		end
	end
end

function SpellReminder:tmerge(from,to)   -- "to" must be a table (possibly empty)
	assert(type(from)=="table","Error merging table contents, source is not a table.")
	assert(type(to)  =="table","Error merging table contents, destination is not a table.")
	for k,v in pairs(from) do
		if(type(v)=="table") then
			if (type(to[k])~='table') then to[k] = {} end
			self:tmerge(v,to[k])
		else
			to[k] = v
		end
	end
end